/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */

#include "config.h"
#include <X11/Xos.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/FileSB.h>
#include <Xm/Frame.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/LabelG.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/Label.h>
#include <Xm/ScrollBar.h>
#include <Xm/DrawnB.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/param.h>
#include "xmap.h"
#include "mode.h"
#include "debug.h"

extern XtAppContext app;
extern Widget status_1, status_2, status_3, draww;

static Position x, y;

struct mapdata mapdata;

enum {
    UP, DOWN, DOWN1,
};
static int state;
static XtIntervalId id;
static Position locx, locy;

Time clktime;

char *a_dist_bear(int x0, int y0, int x1, int y1);

static int routing();
static int rison;

#define TH 3

void update_ini(struct mapdata *mapdata)
{
    char buf1[255];
    char buf2[1024];
    char bufp1[21];
    char bufp2[21];
    struct map *map = mapdata->map;
    int i;

    i = strlen(map->desname);
    if (!strcmp(map->desname + (i - 4), ".des")) {
      write_des(map);
    } else {
      if (!strcmp(map->desname + (i - 4), ".ini")) {
	snprintf(buf2, 1023, "%s/%s", get_mapdir(), map->desname);
	WritePrivateProfileString("general", "maptitle", map->description, 
				  buf2);
	WritePrivateProfileString("general", "filename", map->filename, 
				  buf2);
	switch (mapdata->type) {
	case SRASTER:
	  WritePrivateProfileString("general", "format", "xpm", 
				    buf2);
	  break;
	case DRG:
	  WritePrivateProfileString("general", "format", "tiff", 
				    buf2);
	  break;
	case BSB:
	  WritePrivateProfileString("general", "format", "bsb", 
				    buf2);
	  break;
	case VECTOR:
	  WritePrivateProfileString("general", "format", "vector", 
				    buf2);
	  break;
	case SHAPE:
	  WritePrivateProfileString("general", "format", "shape", 
				    buf2);
	  break;
	}
	toRawLatDMM(mapdata->lat1, bufp1, 21);
	toRawLonDMM(mapdata->lon1, bufp2, 21);
	sprintf(buf1, "%s%s, %d, %d", bufp1, bufp2, mapdata->mx1, 
		mapdata->my1);
	WritePrivateProfileString("georef", "ref1", buf1, 
				  buf2);
	toRawLatDMM(mapdata->lat2, bufp1, 21);
	toRawLonDMM(mapdata->lon2, bufp2, 21);
	sprintf(buf1, "%s%s, %d, %d", bufp1, bufp2, mapdata->mx2, 
		mapdata->my2);
	WritePrivateProfileString("georef", "ref2", buf1, 
				  buf2);
	WritePrivateProfileString("neighbors", "north", map->nmap,
				  buf2);
	WritePrivateProfileString("neighbors", "northeast", map->nemap,
				  buf2);
	WritePrivateProfileString("neighbors", "northwest", map->nwmap,
				  buf2);
	WritePrivateProfileString("neighbors", "east", map->emap,
				  buf2);
	WritePrivateProfileString("neighbors", "southeast", map->semap,
				  buf2);
	WritePrivateProfileString("neighbors", "west", map->wmap,
				  buf2);
	WritePrivateProfileString("neighbors", "southwest", map->swmap,
				  buf2);
	WritePrivateProfileString("neighbors", "south", map->smap,
				  buf2);
	WritePrivateProfileFloat("edges", "north", map->nedge,
				  buf2);
	WritePrivateProfileFloat("edges", "south", map->sedge,
				  buf2);
	WritePrivateProfileFloat("edges", "west", map->wedge,
				  buf2);
	WritePrivateProfileFloat("edges", "east", map->eedge,
				  buf2);
      }
    }
}

void write_des(struct map *map)
{
    FILE *fp;
    char bufp[256];
    char bufp1[21];
    char bufp2[21];

    snprintf(bufp, 255, "%s/%s", get_mapdir(), map->desname);
    if ( (fp = fopen(bufp, "w")) ) {
	fprintf(fp, "%s\n", map->description);
	fprintf(fp, "%s\n", map->filename);
	fprintf(fp, "%s\n", map->ll);
	toRawLatDMM(mapdata.lat1, bufp1, 21);
	toRawLonDMM(mapdata.lon1, bufp2, 21);
	fprintf(fp, "%s%s, %d, %d\n", bufp1, bufp2, mapdata.mx1, mapdata.my1);
	toRawLatDMM(mapdata.lat2, bufp1, 21);
	toRawLonDMM(mapdata.lon2, bufp2, 21);
	fprintf(fp, "%s%s, %d, %d\n", bufp1, bufp2, mapdata.mx2, mapdata.my2);
	fprintf(fp, "%s\n", map->waw);
	fclose(fp);
    } else
      fprintf(stderr, "could not open %s\n", bufp); /*FIXME use error dialog*/
}

static void to_cb(Widget w)
{
    char lat[20], lon[20];
    double dlat, dlon;
    char bufp[256];

    id = 0;

    if (mapdata.calibrated) {
	xytolatlon(locx, locy, &dlat, &dlon);
	if (options.format == UTM) {
	    double dx, dy;
	    char zone[8];

	    DegToUTM(dlat, dlon, zone, &dx, &dy);
	    if (state == DOWN1 || routing()) {
		snprintf(bufp, 255, "%.0fE %.0fN Zone: %s  %s", dx, dy, zone,
			a_dist_bear(x, y, locx, locy));
	    } else
		snprintf(bufp, 255, "%.0fE %.0fN Zone: %s", dx, dy, zone);
	} else {
	    toUserFormatLat(options.format, dlat, lat, 20);
	    toUserFormatLon(options.format, dlon, lon, 20);
	    if (state == DOWN1 || routing()) {
		snprintf(bufp, 255, "%s %s %s",
			lat, lon, a_dist_bear(x, y, locx, locy));
	    } else
		snprintf(bufp, 255, "%s %s", lat, lon);
	}
    } else
	snprintf(bufp, 255, "x=%d y=%d",
		locx * mapdata.zoom + mapdata.mx_offset,
		locy * mapdata.zoom + mapdata.my_offset);
    XmTextFieldSetString(status_1, bufp);
}

int routing()
{
    rison = rison && mode_is(CROUTE);
    return rison;
}


void draw_motion(Widget w, XMotionEvent * e)
{
    locx = e->x;
    locy = e->y;

#if 0
    // This kills the 1 second timer! WHY?
    if (id) {
	XtRemoveTimeOut(id);
	id = 0;
    }
    id = XtAppAddTimeOut(app, 50L, (XtTimerCallbackProc) to_cb, w);
#else
    to_cb(w);
#endif

    if (mode_is(MARK | CROUTE)) {
	mark_move(e->x, e->y);
    } else if (mode_is(NAVIGATE) && state == DOWN &&
	       ((ABSDIFF(x, e->x) > TH && ABSDIFF(y, e->y) > TH) ||
		is_box_on()) && !routing()) {
	undraw_box(w);
	draw_box(w, x, y, e->x, e->y);
    } else if (mode_is(NAVIGATE) && state == DOWN1) {
	undraw_rubber(w);
	draw_rubber(w, x, y, e->x, e->y);
    }
}

void draw_up(Widget w, XMotionEvent * e)
{

    if (mode_is(CALIBRATE)) {
	state = UP;
	if (mapdata.calstate == 0) {
	    if (lldialog("Enter Lat/Lon for Point 1", mapdata.lat1, mapdata.lon1) != ANS_OK) {
		return;
	    }
	    mapdata.mx1 = e->x * mapdata.zoom + mapdata.mx_offset;
	    mapdata.my1 = e->y * mapdata.zoom + mapdata.my_offset;
	    mapdata.lat1 = getlat();
	    mapdata.lon1 = getlon();
	} else if (mapdata.calstate == 1) {
	    if (lldialog("Enter Lat/Lon for Point 2", mapdata.lat2, mapdata.lon2) != ANS_OK) {
		return;
	    }
	    mapdata.mx2 = e->x * mapdata.zoom + mapdata.mx_offset;
	    mapdata.my2 = e->y * mapdata.zoom + mapdata.my_offset;
	    mapdata.lat2 = getlat();
	    mapdata.lon2 = getlon();
	}
    } else if (mode_is(MARK)) {
	mark(e->x, e->y);
    } else if (ABSDIFF(x, e->x) > TH && ABSDIFF(y, e->y) > TH
	       && state == DOWN && !routing()) {

	//Select is done, do something now!

	if (mapdata.type == VECTOR) {
	    set_cursor(True);
	    change_vector(x, y, e->x, e->y);
	    set_cursor(False);
	}
	else if (mapdata.type == BSB || mapdata.type == DRG) {
	    set_cursor(True);
	    change_raster(x, y, e->x, e->y);
	    set_cursor(False);
	}
	undraw_box(w);
	redraw_mark(draww, NULL, NULL);
	redraw_objects();
	redraw_location();
    } else if (state == DOWN1) {
	if (mode_is(CROUTE)) {
	    rison = 1;
	    route(e->x, e->y);
	} else {
	    undraw_rubber(w);	// we dont need it any more.

	    if (e->x == x && e->y == y) {
		beep();		// .
	    } else {
		route_change(x, y, e->x, e->y);
		redraw_mark(draww, NULL, NULL);
		redraw_objects();
		redraw_location();
	    }
	}
    } else {
	if (!place_object(e->x, e->y)) {
	    if (routing()) undraw_rubber(w);
	    center(e->x, e->y);
	    if (routing()) route_translate_rubber(e->x, e->y);
	}
    }
    state = UP;
}

void draw_down(Widget w, XMotionEvent * e)
{
    clktime = e->time;
    x = e->x;
    y = e->y;
    state = DOWN;
}

void draw_enter(Widget w, XMotionEvent * e)
{
}

void draw_leave(Widget w, XMotionEvent * e)
{
}

void draw_select(Widget w, XMotionEvent * e)
{
    clktime = e->time;
    x = e->x;
    y = e->y;
    state = DOWN1;
}


int do_calibrate(void)
{
    double dlat, dlon;
    int dlati, dloni;

// fprintf(stderr,"calibrate: lat1=%f lat2=%f\n",mapdata.lat1, mapdata.lat2);
// fprintf(stderr,"calibrate: lon1=%f lon2=%f\n",mapdata.lon1, mapdata.lon2);


    dlat = ABSDIFF(mapdata.lat1, mapdata.lat2);
    dlon = ABSDIFF(mapdata.lon1, mapdata.lon2);

    dloni = ABSDIFF(mapdata.mx1, mapdata.mx2);
    dlati = ABSDIFF(mapdata.my1, mapdata.my2);

    if (dlat<0.001 || dlon<0.001 || dloni==0 || dlati==0)
	return 0;

    mapdata.pixd_y = (double) dlati / dlat;
    mapdata.pixd_x = (double) dloni / dlon;

    mapdata.nulx = mapdata.lon1 - ((double) mapdata.mx1 / mapdata.pixd_x);
    mapdata.nuly = mapdata.lat1 + ((double) mapdata.my1 / mapdata.pixd_y);

    mapdata.calibrated = 1;

    draw_scale();

    return 1;
}

/* 
 * Read Lat/Lon and convert from HHMMDD and HHHMMDD to DD.DDDDDD and DDD.DDDDDD
 *
 */

void init_map(struct map *map)
{
    double lat1, lon1;
    double lat2, lon2;
    double d, m;
    long lat, lon;
    char p, q;

    mapdata.mx_offset = 0;
    mapdata.my_offset = 0;


    sscanf(map->ref1, "%ld%c%ld%c, %d, %d", &lat, &p, &lon, &q,
	   &(mapdata.mx1), &(mapdata.my1));

    lat1 = (double)lat/100.0;
//    lat1 /= 100.0;
    m = 100.0 * modf(lat1 / 100.0, &d);
    lat1 = d + m / 60.0;

    lon1 = (double)lon/100.0;
//    lon1 /= 100.0;
    m = 100.0 * modf(lon1 / 100.0, &d);
    lon1 = d + m / 60.0;

    if (p == 'S')
	lat1 = -lat1;
    if (q == 'W')
	lon1 = -lon1;




    sscanf(map->ref2, "%ld%c%ld%c, %d, %d", &lat, &p, &lon, &q,
	   &(mapdata.mx2), &(mapdata.my2));

    if (mapdata.mx1 == 0 && mapdata.my1 == 0
	&& mapdata.mx2 == 0 && mapdata.my2 == 0
	&& !mapdata.calibrated) {
	XmTextFieldSetString(status_1,
			     "Uncalibrated map, x/y display only.");
	return;
    }

    lat2 = (double)lat/100.0;
//    lat2 /= 100.0;
    m = 100.0 * modf(lat2 / 100.0, &d);
    lat2 = d + m / 60.0;

    lon2 = (double)lon/100.0;
//    lon2 /= 100.0;
    m = 100.0 * modf(lon2 / 100.0, &d);
    lon2 = d + m / 60.0;

    if (p == 'S')
	lat2 = -lat2;
    if (q == 'W')
	lon2 = -lon2;

    mapdata.lat1 = lat1;
    mapdata.lon1 = lon1;
    mapdata.lat2 = lat2;
    mapdata.lon2 = lon2;

    do_calibrate();
}

void vector_map(struct map *map)
{
    double lat1, lon1;
    double lat2, lon2;
    double d, m;

    char p, q;

    sscanf(map->ref1, "%lf%c%lf%c, %d, %d", &lat1, &p, &lon1, &q,
	   &(mapdata.mx1), &(mapdata.my1));

    lat1 /= 100.0;
    m = 100.0 * modf(lat1 / 100.0, &d);
    lat1 = d + m / 60.0;

    lon1 /= 100.0;
    m = 100.0 * modf(lon1 / 100.0, &d);
    lon1 = d + m / 60.0;

    if (p == 'S')
	lat1 = -lat1;
    if (q == 'W')
	lon1 = -lon1;


    sscanf(map->ref2, "%lf%c%lf%c, %d, %d", &lat2, &p, &lon2, &q,
	   &(mapdata.mx2), &(mapdata.my2));

    lat2 /= 100.0;
    m = 100.0 * modf(lat2 / 100.0, &d);
    lat2 = d + m / 60.0;

    lon2 /= 100.0;
    m = 100.0 * modf(lon2 / 100.0, &d);
    lon2 = d + m / 60.0;

    if (p == 'S')
	lat2 = -lat2;
    if (q == 'W')
	lon2 = -lon2;

    mapdata.lat1 = lat1;
    mapdata.lon1 = lon1;
    mapdata.lat2 = lat2;
    mapdata.lon2 = lon2;

    mapdata.calibrated = 1;
}



int lon2x(double lon)
{
    return ((int) ((lon - mapdata.nulx) * mapdata.pixd_x) - mapdata.mx_offset) / mapdata.zoom;
}

int lat2y(double lat)
{
    return (-(int) ((lat - mapdata.nuly) * mapdata.pixd_y) - mapdata.my_offset) / mapdata.zoom;
}

double x2lon(int x)
{
    return mapdata.nulx + (double) (x * mapdata.zoom + mapdata.mx_offset) / mapdata.pixd_x;
}

double y2lat(int y)
{
    return mapdata.nuly - (double) (y * mapdata.zoom + mapdata.my_offset) / mapdata.pixd_y;
}

void latlontoxy(double lat, double lon, Position * x, Position * y)
{
    datum_translate(1, &lat, &lon, mapdata.mapdatum);

    switch (mapdata.type) {
	case VECTOR:
	    project(lat, lon, x, y);
	    break;
	case BSB:
	    bsb_project(lat, lon, x, y);
	    break;
	case DRG:
	    gtif_project(lat, lon, x, y);
	    break;
	default:
	    *x = lon2x(lon);
	    *y = lat2y(lat);
	    break;
    }
}

void xytolatlon(Position x, Position y, double *lat, double *lon)
{
    switch (mapdata.type) {
	case VECTOR:
	    decode(x, y, lat, lon);
	    break;
	case BSB:
	    bsb_decode(x, y, lat, lon);
	    break;
	case DRG:
	    gtif_decode(x, y, lat, lon);
	    break;
	default:
	    *lon = x2lon(x);
	    *lat = y2lat(y);
	    break;
    }
    datum_translate(0, lat, lon, mapdata.mapdatum);
}

double distance(double lata, double lona, double latb, double lonb)
{
 /*_________________________________________________________________________
 ** Grosskreis - Distanz, FAI - Berechnung
 ** gem. Code Sportif 1992
 ** Absatz 2.11.4 Seite 2-8 unterste Formel
 ** 1 Grad = 111.1949267 km
 */

    double l0 = lona * DEGREE;
    double l1 = lonb * DEGREE;
    double b0 = lata * DEGREE;
    double b1 = latb * DEGREE;

    return (6371 * 2 * asin(sqrt(cos(b1) * cos(b0) * sin(0.5 * (l1 - l0))
				 * sin(0.5 * (l1 - l0))
			+ sin(0.5 * (b1 - b0)) * sin(0.5 * (b1 - b0)))));
}

double heading(double lata, double lona, double latb, double lonb)
{
    double tt = 0.0;
    double dlat, dlong;
    int offset, factor;

#if 0
    double lat1, lon1, lat2, lon2, lat3, lon3;

    lat1 = lata;
    lon1 = lona;
    lat2 = latb;
    lon2 = lona;
    lat3 = latb;
    lon3 = lonb;
#endif
    dlat = latb - lata;
    dlong = lonb - lona;


    if (dlat == 0.0) {
	if (dlong > (double) 0.0)
	    tt = 90.0;
	if (dlong < 0.0)
	    tt = 270.0;
    } else {
	offset = 0;
	factor = 1;

	if ((dlat < 0.0) && (dlong > 0.0)) {
	    offset = 90;
	    factor = -1;
	}
	if ((dlat < 0.0) && (dlong <= 0.0))
	    offset = 180;
	if ((dlat > 0.0) && (dlong < 0.0)) {
	    offset = 270;
	    factor = -1;
	}
	dlat *= (dlat >= 0.0) ? 1.0 : -1.0;
	dlong *= (dlong >= 0.0) ? 1.0 : -1.0;

#if 1
	tt = (double) factor *(180 * atan(dlong / dlat) / PI);
#else
	tt = (double) factor *(180 * atan(distance(lat2, lon2, lat3, lon3) /
				 distance(lat1, lon1, lat2, lon2)) / PI);
#endif

	if (tt < 0.0)
	    tt += 90.0;
	tt += (double) offset;
    }

    return tt;
}

char *a_dist_bear(int x0, int y0, int x1, int y1)
{
    double lata, lona, latb, lonb;
    double d, h;
    char *p;
    static char bufp[20];

    xytolatlon(x0, y0, &lata, &lona);
    xytolatlon(x1, y1, &latb, &lonb);
    d = cv_dist(distance(lata, lona, latb, lonb));
    h = heading(lata, lona, latb, lonb);

    p = "Km";

    switch (unit) {
//    case METRIC:
//	p = "Km";
//	break;
    case NAUTICAL:
	p = "nm";
	break;
    case ENGLISH:
	p = "mi";
	break;
    }
    snprintf(bufp, 19, "  %.1f %s  %03.0f", d, p, h);
    return bufp;
}
