/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */

#include "config.h"
#include <stdio.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <errno.h>
#include "gps.h"
#include "debug.h"

extern XtAppContext app;
extern Widget draw_w;
static Widget dialog;
static void dlg_callback();
static int err_ack;

static void create_dialog()
{
    Arg args[5];
    int n = 0;
    XmString ok = XmStringCreateLocalized("OK");

    XtSetArg(args[n], XmNautoUnmanage, False);
    n++;
    XtSetArg(args[n], XmNcancelLabelString, ok);
    n++;
    XtSetArg(args[n], XmNhighlightThickness, 0);
    n++;
    XtSetArg(args[n], XmNtitle, "Error");
    n++;
    dialog = XmCreateErrorDialog(draw_w, "Error", args, n);
    XtAddCallback(dialog, XmNcancelCallback, dlg_callback, NULL);
    XtUnmanageChild( XmMessageBoxGetChild(dialog, XmDIALOG_OK_BUTTON));
    XtUnmanageChild( XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
}

void error(char *errmsg, int Errno)
{
    XmString t;
    char bufp[1024];

    if (!dialog)
	create_dialog();

    if (Errno) {
	snprintf(bufp, 1023, "%s\n%s", errmsg, strerror(Errno));
    }
    else
	snprintf(bufp, 1023, "%s", errmsg);

    t = XmStringCreateLtoR(bufp, XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(dialog,
		  XmNmessageString, t,
		  XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
		  NULL);
    XmStringFree(t);

    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);
    err_ack = 0;
}

void internal_error(char *errmsg)
{
    XmString t;
    char bufp[1024];

    if (!dialog)
	create_dialog();

    snprintf(bufp, 1023, "%s\nPlease report this error!", errmsg);

    t = XmStringCreateLtoR(bufp, XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(dialog,
		  XmNmessageString, t,
		  XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
		  NULL);
    XmStringFree(t);

    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);
}


void errexit(char *errmsg, int Errno)
{
    error(errmsg, Errno);
    io_error();
}

static void dlg_callback(Widget dialog, XtPointer client_data, XtPointer call_data)
{
    XtPopdown(XtParent(dialog));
    err_ack = 1;
}

void error_wait_ack(void)
{
    while(!err_ack)
	XtAppProcessEvent(app, XtIMAll);
}

void error_wait(char *errmsg, int Errno)
{
    error(errmsg, Errno);
    error_wait_ack();
    exit(1);
}
