#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>

#include <X11/Xos.h>
#include <X11/cursorfont.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/FileSB.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/LabelG.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/ScrollBar.h>
#include <X11/xpm.h>

#include "xmap.h"
#include "mode.h"
#include "debug.h"

extern Widget toplevel;
extern Widget draww;
extern GC clearGC;

static struct map *lmap;
static Dimension lwidth, lheight;

#define PROJ 0

void vprint()
{
    ImageFilledRectangle(NULL, 0, 0, lmap->mwidth, lmap->mheight, clearGC);
    DrawVectorMap(lmap->pmap, lmap->mwidth, lmap->mheight,
		  mapdata.clat, mapdata.clon, mapdata.wlon, mapdata.wlat, PROJ);
}

static void vdraw()
{
    Display *dpy = XtDisplay(draww);

    XFillRectangle(dpy, lmap->pmap, clearGC, 0, 0, lmap->mwidth, lmap->mheight);
    DrawVectorMap(lmap->pmap, lmap->mwidth, lmap->mheight,
		  mapdata.clat, mapdata.clon, mapdata.wlon, mapdata.wlat, PROJ);
    refresh(0);
    draw_scale();
}

do_vector(struct map *map)
{
    Pixmap new;
    XColor color;
    Colormap cmap = DefaultColormapOfScreen(XtScreen(draww));
    Display *dpy = XtDisplay(draww);

    lmap = map;

    XtVaGetValues(draww,
		  XmNwidth, &(lmap->mwidth),
		  XmNheight, &(lmap->mheight),
		  NULL);

    lwidth = lmap->mwidth;
    lheight = lmap->mheight;

    new = XCreatePixmap(dpy, XtWindow(draww), lmap->mwidth,
		   lmap->mheight, DefaultDepthOfScreen(XtScreen(draww)));
    color.red = (214 << 8);
    color.green = (227 << 8);
    color.blue = (255 << 8);
    myXAllocColor(dpy, cmap, &color);
    XSetForeground(dpy, clearGC, color.pixel);
    if (lmap->pmap)
	XFreePixmap(dpy, lmap->pmap);
    lmap->pmap = new;
}

cvll(double lat1, double lon1, double lat2, double lon2)
{
    Display *dpy = XtDisplay(draww);

    mapdata.clat = (lat1 + lat2) / 2.0;
    mapdata.clon = (lon1 + lon2) / 2.0;

    mapdata.wlat = ABSDIFF(lat1, lat2);
    mapdata.wlon = ABSDIFF(lon1, lon2);



//fprintf(stderr,"clat=%f clon=%f  wlat=%f wlon=%f\n",mapdata.clat,mapdata.clon,mapdata.wlat,mapdata.wlon);



    XtVaGetValues(draww,
		  XmNwidth, &(lmap->mwidth),
		  XmNheight, &(lmap->mheight),
		  NULL);

    if (lwidth != lmap->mwidth || lheight != lmap->mheight) {
	lwidth = lmap->mwidth;
	lheight = lmap->mheight;
	XFreePixmap(dpy, lmap->pmap);
	lmap->pmap = XCreatePixmap(dpy, XtWindow(draww), lmap->mwidth,
		   lmap->mheight, DefaultDepthOfScreen(XtScreen(draww)));
    }
    vdraw();
}

void change_vector(int x1, int y1, int x2, int y2)
{
    double lat1, lon1, lat2, lon2;

    decode(x1, y1, &lat1, &lon1);
    decode(x2, y2, &lat2, &lon2);

//fprintf(stderr,"lat1=%f lon1=%f  lat2=%f lon2=%f\n",lat1,lon1,lat2,lon2);
    cvll(lat1,lon1,lat2,lon2);
}

void vcenter(int x, int y)
{
    Display *dpy = XtDisplay(draww);

    decode(x, y, &(mapdata.clat), &(mapdata.clon));

//fprintf(stderr,"clat=%f clon=%f  wlat=%f wlon=%f\n",mapdata.clat,mapdata.clon,mapdata.wlat,mapdata.wlon);
    vdraw();
}

void vzoomin()
{
    Display *dpy = XtDisplay(draww);

    if (mapdata.type == VECTOR) {
	set_cursor(True);
	mapdata.wlon /= 2.0;
	mapdata.wlat /= 2.0;

	vdraw();
	set_cursor(False);
    } else
	beep();
}

void vzoomout()
{
    Display *dpy = XtDisplay(draww);

    if (mapdata.type == VECTOR) {
	set_cursor(True);
	mapdata.wlon *= 2.0;
	mapdata.wlat *= 2.0;

	vdraw();
	set_cursor(False);
    } else
	beep();
}

void vzoomout4()
{
    Display *dpy = XtDisplay(draww);

    if (mapdata.type == VECTOR) {
	set_cursor(True);
	mapdata.wlon *= 4.0;
	mapdata.wlat *= 4.0;

	vdraw();
	set_cursor(False);
    } else
	beep();
}
